#!/usr/bin/env python3
"""
Report Card Data Aggregation Module for Report 001

This module takes raw virus data from retrieve_virus_data.py and applies the
aggregation logic from labkey_files/names.sql to create report card data.

The module replicates the SQL logic entirely in pandas without additional
database queries, producing the same data structure as the original SQL.

Usage:
    from create_report_card_data import create_report_card_data
    
    # Get raw virus data
    raw_df = retrieve_all_virus_data()
    
    # Create report card data
    report_card_df = create_report_card_data(raw_df)
"""

import sys
import os
import tomllib
from pathlib import Path
from typing import Dict, List, Any, Optional
import pandas as pd
import numpy as np
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn
from rich.table import Table
import pickle
import gzip
from datetime import datetime

# Load environment variables from .env file if running locally
try:
    from dotenv import load_dotenv
    # Look for .env file in the project root
    env_path = Path(__file__).parent.parent.parent / '.env'
    if env_path.exists():
        load_dotenv(env_path)
except ImportError:
    # dotenv not available, rely on environment variables being set externally
    pass

# Import the virus data retrieval module
from retrieve_virus_data import retrieve_all_virus_data, save_to_cache, load_from_cache


def create_base_data(df: pd.DataFrame) -> pd.DataFrame:
    """
    Create base data matching the 'base' CTE from names.sql.
    
    Args:
        df: Raw virus data from retrieve_virus_data
        
    Returns:
        Filtered and formatted base DataFrame
    """
    if df.empty:
        return pd.DataFrame()
    
    # Apply base filters (matching the SQL WHERE clause)
    base_df = df.copy()
    
    # Filter conditions from names.sql base CTE
    base_df = base_df[
        base_df['sewershed_location'].notna() &
        (base_df['sewershed_location'] != '') &
        base_df['sample_date'].notna()
    ]
    
    # Convert sample_date to datetime and create time_label
    base_df['sample_date'] = pd.to_datetime(base_df['sample_date'], errors='coerce')
    base_df = base_df[base_df['sample_date'].notna()]  # Remove invalid dates
    
    # Create time_label (YYYY-MM-DD format)
    base_df['time_label'] = base_df['sample_date'].dt.strftime('%Y-%m-%d')
    
    # Handle both old (lowercase) and new (capitalized) field names from LabKey
    rank_col = 'Rank' if 'Rank' in base_df.columns else 'rank'
    species_col = 'Species' if 'Species' in base_df.columns else 'species'
    stitle_col = 'Stitle' if 'Stitle' in base_df.columns else 'stitle'
    mapped_reads_col = 'Mapped_Reads' if 'Mapped_Reads' in base_df.columns else 'mapped_reads'
    total_reads_col = 'Total_Reads' if 'Total_Reads' in base_df.columns else 'total_reads'
    
    # Select and reorder columns to match SQL
    columns = ['sewershed_location', 'sample_date', 'time_label', rank_col, species_col, stitle_col, mapped_reads_col, total_reads_col]
    base_df = base_df[columns]
    
    # Standardize column names for consistency in downstream processing
    base_df = base_df.rename(columns={
        rank_col: 'rank',
        species_col: 'species', 
        stitle_col: 'stitle',
        mapped_reads_col: 'mapped_reads',
        total_reads_col: 'total_reads'
    })
    
    return base_df


def load_virus_config(config_path: Optional[Path] = None) -> Dict[str, Any]:
    """
    Load virus configuration from virus.toml file.
    
    Args:
        config_path: Path to virus.toml file. If None, looks in current directory.
        
    Returns:
        Dictionary with virus configuration
    """
    if config_path is None:
        config_path = Path(__file__).parent / "virus.toml"
    else:
        config_path = Path(config_path)
    
    if not config_path.exists():
        raise FileNotFoundError(f"Virus configuration file not found: {config_path}")
    
    try:
        with open(config_path, 'rb') as f:
            config = tomllib.load(f)
        return config
    except Exception as e:
        raise Exception(f"Failed to load virus configuration from {config_path}: {e}")


def apply_virus_search_criteria(mapped_df: pd.DataFrame, search_criteria: List[str]) -> pd.Series:
    """
    Apply virus search criteria and return boolean mask.
    
    Args:
        mapped_df: DataFrame with lowercase columns for matching
        search_criteria: List of SQL LIKE conditions
        
    Returns:
        Boolean series indicating matches
    """
    import re
    
    if not search_criteria:
        return pd.Series([False] * len(mapped_df))
    
    all_conditions = []
    
    for criterion in search_criteria:
        # Handle OR conditions FIRST (parentheses format)
        if 'OR' in criterion.upper():
            # Split on OR and process each part
            or_parts = [part.strip() for part in criterion.split(' OR ')]
            or_conditions = []
            
            for part in or_parts:
                # Remove parentheses if present
                part = part.strip('()')
                
                # Parse this part as a simple LIKE
                part_match = re.search(r'(\w+)\s+LIKE\s+[\'"]([^\'\"]+)[\'"]', part, re.IGNORECASE)
                if part_match:
                    column = part_match.group(1).lower()
                    pattern = part_match.group(2).replace('%', '')
                    
                    if column == 'rank':
                        or_conditions.append(mapped_df['rank_lower'].str.contains(pattern, case=False, na=False))
                    elif column == 'species':
                        or_conditions.append(mapped_df['species_lower'].str.contains(pattern, case=False, na=False))
                    elif column == 'stitle':
                        or_conditions.append(mapped_df['stitle_lower'].str.contains(pattern, case=False, na=False))
            
            if or_conditions:
                # Combine OR conditions
                combined_or = or_conditions[0]
                for cond in or_conditions[1:]:
                    combined_or = combined_or | cond
                all_conditions.append(combined_or)
            continue
        
        # Parse NOT LIKE conditions
        not_like_match = re.search(r'(\w+)\s+NOT\s+LIKE\s+[\'"]([^\'\"]+)[\'"]', criterion, re.IGNORECASE)
        if not_like_match:
            column = not_like_match.group(1).lower()
            pattern = not_like_match.group(2).replace('%', '')
            
            if column == 'rank':
                condition = ~mapped_df['rank_lower'].str.contains(pattern, case=False, na=False)
            elif column == 'species':
                condition = ~mapped_df['species_lower'].str.contains(pattern, case=False, na=False)
            elif column == 'stitle':
                condition = ~mapped_df['stitle_lower'].str.contains(pattern, case=False, na=False)
            else:
                continue
                
            all_conditions.append(condition)
            continue
        
        # Parse simple LIKE conditions (moved to last)
        like_match = re.search(r'(\w+)\s+LIKE\s+[\'"]([^\'\"]+)[\'"]', criterion, re.IGNORECASE)
        if like_match:
            column = like_match.group(1).lower()
            pattern = like_match.group(2).replace('%', '')
            
            if column == 'rank':
                condition = mapped_df['rank_lower'].str.contains(pattern, case=False, na=False)
            elif column == 'species':
                condition = mapped_df['species_lower'].str.contains(pattern, case=False, na=False)
            elif column == 'stitle':
                condition = mapped_df['stitle_lower'].str.contains(pattern, case=False, na=False)
            else:
                continue
                
            all_conditions.append(condition)
            continue
    
    # Combine all conditions with AND
    if not all_conditions:
        return pd.Series([False] * len(mapped_df))
    
    result = all_conditions[0]
    for condition in all_conditions[1:]:
        result = result & condition
    
    return result


def map_virus_names(df: pd.DataFrame) -> pd.DataFrame:
    """
    Map virus names using configuration from virus.toml.
    
    Args:
        df: Base DataFrame
        
    Returns:
        DataFrame with virus names mapped
    """
    if df.empty:
        return pd.DataFrame()
    
    # Load virus configuration
    try:
        config = load_virus_config()
    except Exception as e:
        raise Exception(f"Failed to load virus configuration: {e}")
    
    # Create copy and add virus column
    mapped_df = df.copy()
    
    # Handle both old (lowercase) and new (capitalized) field names from LabKey
    rank_col = 'Rank' if 'Rank' in mapped_df.columns else 'rank'
    species_col = 'Species' if 'Species' in mapped_df.columns else 'species'
    stitle_col = 'Stitle' if 'Stitle' in mapped_df.columns else 'stitle'
    
    # Convert to lowercase for matching (like SQL LOWER())
    mapped_df['rank_lower'] = mapped_df[rank_col].str.lower().fillna('')
    mapped_df['species_lower'] = mapped_df[species_col].str.lower().fillna('')
    mapped_df['stitle_lower'] = mapped_df[stitle_col].str.lower().fillna('')
    
    # Initialize virus column
    mapped_df['virus'] = None
    
    # Apply virus mapping from virus.toml
    virus_section = config.get('virus', {})
    
    for virus_data in virus_section.values():
        virus_name = virus_data.get('name')
        search_criteria = virus_data.get('search_criteria', [])
        
        if not virus_name or not search_criteria:
            continue
        
        try:
            # Apply search criteria to get matching rows
            matches = apply_virus_search_criteria(mapped_df, search_criteria)
            
            # Set virus name for matching rows (only if not already set - first match wins)
            mapped_df.loc[matches & mapped_df['virus'].isna(), 'virus'] = virus_name
            
        except Exception as e:
            # Skip viruses that fail to process
            continue
    
    # Filter to only rows with virus assignment (WHERE virus IS NOT NULL)
    virus_assigned = mapped_df[mapped_df['virus'].notna()].copy()
    
    # Select columns for virus_assigned
    virus_assigned = virus_assigned[['sewershed_location', 'time_label', 'virus', 'mapped_reads', 'total_reads']]
    
    # Clean up temporary columns
    mapped_df.drop(['rank_lower', 'species_lower', 'stitle_lower'], axis=1, inplace=True, errors='ignore')
    
    return virus_assigned


def create_additional_totals(base_df: pd.DataFrame) -> List[pd.DataFrame]:
    """
    Create additional total categories from virus.toml configuration.
    
    Args:
        base_df: Base DataFrame
        
    Returns:
        List of DataFrames for total categories defined in virus.toml
    """
    if base_df.empty:
        return []
    
    # Load virus configuration
    try:
        config = load_virus_config()
    except Exception as e:
        raise Exception(f"Failed to load virus configuration: {e}")
    
    # Convert to lowercase for matching
    base_lower = base_df.copy()
    base_lower['rank_lower'] = base_lower['rank'].str.lower().fillna('')
    base_lower['species_lower'] = base_lower['species'].str.lower().fillna('')
    base_lower['stitle_lower'] = base_lower['stitle'].str.lower().fillna('')
    
    total_dataframes = []
    virus_section = config.get('virus', {})
    
    # Find viruses that are "total" categories
    for virus_data in virus_section.values():
        virus_name = virus_data.get('name', '')
        search_criteria = virus_data.get('search_criteria', [])
        
        # Check if this is a total category (name contains "total" or similar patterns)
        if '(total)' in virus_name.lower() or 'total' in virus_name.lower():
            try:
                # Apply search criteria to find matching rows
                matches = apply_virus_search_criteria(base_lower, search_criteria)
                matching_rows = base_lower[matches].copy()
                
                if not matching_rows.empty:
                    matching_rows['virus'] = virus_name
                    matching_rows = matching_rows[['sewershed_location', 'time_label', 'virus', 'mapped_reads', 'total_reads']]
                    total_dataframes.append(matching_rows)
            except Exception:
                continue
    
    return total_dataframes


def assign_virus_types(df: pd.DataFrame) -> pd.DataFrame:
    """
    Assign virus types based on virus names using virus.toml configuration.
    
    Args:
        df: DataFrame with virus names
        
    Returns:
        DataFrame with virus_type column added
    """
    if df.empty:
        return df
    
    # Load virus configuration
    try:
        config = load_virus_config()
    except Exception as e:
        raise Exception(f"Failed to load virus configuration: {e}")
    
    result_df = df.copy()
    
    # Build virus type mappings from virus.toml
    virus_section = config.get('virus', {})
    virus_type_mapping = {}
    
    for virus_data in virus_section.values():
        virus_name = virus_data.get('name')
        virus_type = virus_data.get('type')
        
        if virus_name and virus_type:
            virus_type_mapping[virus_name] = virus_type
    
    # Create virus type to virus names mapping
    type_to_viruses = {}
    for virus_name, virus_type in virus_type_mapping.items():
        if virus_type not in type_to_viruses:
            type_to_viruses[virus_type] = set()
        type_to_viruses[virus_type].add(virus_name)
    
    # Assign virus types using configuration
    conditions = []
    choices = []
    
    for virus_type, virus_names in type_to_viruses.items():
        conditions.append(result_df['virus'].isin(virus_names))
        choices.append(virus_type)
    
    if conditions:
        result_df['virus_type'] = np.select(conditions, choices, default=None)
    else:
        result_df['virus_type'] = None
    
    return result_df


def create_report_card_data(raw_df: pd.DataFrame, 
                           test_filter: bool = False,
                           show_progress: bool = True,
                           cache_file: Optional[Path] = None,
                           use_cache: bool = False) -> pd.DataFrame:
    """
    Create report card data from raw virus data, replicating the logic from labkey_files/names.sql.
    
    This function applies the same aggregation and virus mapping logic as the SQL query
    but operates entirely on pandas DataFrames without additional database queries.
    
    Args:
        raw_df: Raw virus data from retrieve_virus_data
        test_filter: If True, filter to Columbia, MO data from June 2024 for fast testing
        show_progress: Whether to show progress indicators
        cache_file: Path to cache file for saving/loading processed data
        use_cache: If True and cache exists, load from cache instead of processing
        
    Returns:
        pandas DataFrame with report card structure:
        - sewershed_location
        - time_label (YYYY-MM-DD)
        - virus (mapped virus name)
        - mapped_reads (summed)
        - total_reads (max)
        - virus_type (enteric/respiratory/atypical/animal)
    """
    console = Console()
    
    # Try to load from cache if requested
    if use_cache and cache_file and cache_file.exists():
        if show_progress:
            console.print(f"[cyan]Loading report card data from cache: {cache_file}[/cyan]")
        
        try:
            df, cache_info = load_from_cache(cache_file)
            age_days = (datetime.now() - cache_info['timestamp']).days
            
            if show_progress:
                console.print(f"[green]✓ Loaded {len(df)} report card records from cache[/green]")
                console.print(f"  Cache age: {age_days} days old")
                console.print(f"  Created: {cache_info['timestamp'].strftime('%Y-%m-%d %H:%M:%S')}")
                
                if age_days > 7:
                    console.print(f"[yellow]⚠ Cache is {age_days} days old. Consider refreshing.[/yellow]")
            
            return df
            
        except Exception as e:
            console.print(f"[yellow]Failed to load cache: {e}[/yellow]")
            console.print("[cyan]Falling back to fresh data processing...[/cyan]")
    
    if raw_df.empty:
        console.print("[yellow]Warning: Empty input DataFrame[/yellow]")
        return pd.DataFrame()
    
    if show_progress:
        console.print("[blue]Creating report card data from raw virus data...[/blue]")
    
    try:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            console=console,
            disable=not show_progress
        ) as progress:
            
            main_task = progress.add_task("Processing virus data...", total=100)
            
            # Step 1: Create base data (equivalent to 'base' CTE)
            progress.update(main_task, description="Creating base data...", completed=10)
            base_df = create_base_data(raw_df)
            
            if base_df.empty:
                console.print("[yellow]Warning: No data after base filtering[/yellow]")
                return pd.DataFrame()
            
            # Apply test filter if requested
            if test_filter:
                progress.update(main_task, description="Applying test filter (Columbia, MO June 2024)...", completed=15)
                base_df = base_df[
                    (base_df['sewershed_location'] == 'Columbia, MO') &
                    (base_df['sample_date'].dt.year == 2024) &
                    (base_df['sample_date'].dt.month == 6)
                ]
                
                if base_df.empty:
                    console.print("[yellow]Warning: No data for Columbia, MO June 2024[/yellow]")
                    return pd.DataFrame()
                
                if show_progress:
                    console.print(f"[cyan]Test filter applied: {len(base_df)} records for Columbia, MO June 2024[/cyan]")
            
            # Step 2: Map virus names (equivalent to 'mapped_viruses' CTE)
            progress.update(main_task, description="Mapping virus names...", completed=30)
            virus_assigned = map_virus_names(base_df)
            
            # Step 3: Create additional totals (influenza_total, adenovirusF_total, rhinovirus_total CTEs)
            progress.update(main_task, description="Creating additional virus totals...", completed=50)
            additional_totals = create_additional_totals(base_df)
            
            # Step 4: Combine all data (equivalent to 'combined' CTE)
            progress.update(main_task, description="Combining all virus data...", completed=70)
            
            all_dataframes = [virus_assigned]
            all_dataframes.extend(additional_totals)
            
            if len(all_dataframes) == 1:
                combined_df = all_dataframes[0]
            else:
                combined_df = pd.concat(all_dataframes, ignore_index=True)
            
            # Step 5: Aggregate data (equivalent to the final GROUP BY)
            progress.update(main_task, description="Aggregating data by location, date, and virus...", completed=85)
            
            # Group by and aggregate (matching the final SQL SELECT)
            final_df = combined_df.groupby(['sewershed_location', 'time_label', 'virus']).agg({
                'mapped_reads': 'sum',  # SUM(mapped_reads)
                'total_reads': 'max'    # MAX(total_reads)
            }).reset_index()
            
            # Step 6: Assign virus types
            progress.update(main_task, description="Assigning virus types...", completed=95)
            final_df = assign_virus_types(final_df)
            
            # Step 7: Final ordering (matching SQL ORDER BY)
            final_df = final_df.sort_values(['sewershed_location', 'time_label', 'virus']).reset_index(drop=True)
            
            progress.update(main_task, description="Report card data creation complete", completed=100)
        
        if show_progress:
            console.print(f"[green]✓ Successfully created report card data: {len(final_df)} records[/green]")
            
            # Show summary statistics
            if not final_df.empty:
                unique_locations = final_df['sewershed_location'].nunique()
                unique_viruses = final_df['virus'].nunique()
                date_range = f"{final_df['time_label'].min()} to {final_df['time_label'].max()}"
                
                console.print(f"[cyan]Summary: {unique_locations} locations, {unique_viruses} viruses, dates {date_range}[/cyan]")
        
        # Save to cache if cache_file is specified
        if cache_file and not final_df.empty:
            if show_progress:
                console.print(f"[cyan]Saving report card data to cache: {cache_file}[/cyan]")
            
            metadata = {
                'test_filter': test_filter,
                'row_count': len(final_df),
                'unique_locations': unique_locations,
                'unique_viruses': unique_viruses,
                'date_range': date_range,
                'columns': list(final_df.columns)
            }
            
            try:
                save_to_cache(final_df, cache_file, metadata)
                if show_progress:
                    console.print(f"[green]✓ Saved {len(final_df)} records to cache[/green]")
            except Exception as e:
                console.print(f"[yellow]Failed to save cache: {e}[/yellow]")
        
        return final_df
        
    except Exception as e:
        error_msg = f"Failed to create report card data: {e}"
        if show_progress:
            console.print(f"[red]✗ {error_msg}[/red]")
        raise Exception(error_msg)


def get_report_card_summary(df: pd.DataFrame) -> Dict[str, Any]:
    """
    Get summary statistics for report card data.
    
    Args:
        df: Report card DataFrame
        
    Returns:
        Dictionary with summary statistics
    """
    if df.empty:
        return {"total_records": 0, "message": "No data available"}
    
    summary = {
        "total_records": len(df),
        "unique_locations": df['sewershed_location'].nunique(),
        "unique_viruses": df['virus'].nunique(),
        "date_range": {
            "earliest": df['time_label'].min(),
            "latest": df['time_label'].max()
        },
        "virus_types": df['virus_type'].value_counts().to_dict(),
        "total_mapped_reads": df['mapped_reads'].sum(),
        "locations": sorted(df['sewershed_location'].unique().tolist()),
        "viruses_by_type": {
            virus_type: sorted(df[df['virus_type'] == virus_type]['virus'].unique().tolist())
            for virus_type in df['virus_type'].dropna().unique()
        }
    }
    
    return summary


def main():
    """Demonstration of report card data creation functionality."""
    console = Console()
    
    console.print("[bold green]📊 Report Card Data Creation Module for Report 001[/bold green]\n")
    
    # Check if API key is available
    if not os.environ.get('LABKEY_API'):
        console.print("[yellow]⚠ LABKEY_API environment variable not set[/yellow]")
        console.print("Set your LabKey API key to test data retrieval:")
        console.print("export LABKEY_API='your_api_key_here'")
        return 1
    
    try:
        # Test 1: Fast test with Columbia, MO June 2024 data
        console.print("[blue]Test 1: Fast test with Columbia, MO June 2024 data...[/blue]")
        
        # Get raw data with test filter applied in retrieve_virus_data
        raw_data = retrieve_all_virus_data(max_rows=1000, show_progress=False)
        
        if raw_data.empty:
            console.print("[red]No raw data available for testing[/red]")
            return 1
        
        # Create report card data with test filter
        test_report_data = create_report_card_data(raw_data, test_filter=True, show_progress=True)
        
        if not test_report_data.empty:
            console.print(f"✓ Test data: {len(test_report_data)} report card records")
            
            # Show sample data
            console.print("\n[cyan]Sample test data (first 3 rows):[/cyan]")
            sample_table = Table()
            for col in test_report_data.columns:
                sample_table.add_column(col, style="white")
            
            for _, row in test_report_data.head(3).iterrows():
                sample_table.add_row(*[str(val) for val in row])
            
            console.print(sample_table)
        else:
            console.print("[yellow]No test data found for Columbia, MO June 2024[/yellow]")
        
        # Test 2: Full processing test (small sample)
        console.print(f"\n[blue]Test 2: Full processing test (100 records sample)...[/blue]")
        
        small_sample = raw_data.head(100)
        full_report_data = create_report_card_data(small_sample, test_filter=False, show_progress=True)
        
        console.print(f"✓ Full processing: {len(full_report_data)} report card records from {len(small_sample)} raw records")
        
        # Test 3: Summary statistics
        if not full_report_data.empty:
            console.print(f"\n[blue]Test 3: Report card data summary...[/blue]")
            summary = get_report_card_summary(full_report_data)
            
            summary_table = Table(title="Report Card Data Summary")
            summary_table.add_column("Property", style="cyan")
            summary_table.add_column("Value", style="white")
            
            for key, value in summary.items():
                if key not in ["locations", "viruses_by_type"]:
                    summary_table.add_row(str(key), str(value))
            
            console.print(summary_table)
            
            # Show virus types breakdown
            if "virus_types" in summary:
                console.print(f"\n[cyan]Virus types found:[/cyan]")
                for virus_type, count in summary["virus_types"].items():
                    console.print(f"  • {virus_type}: {count} records")
        
        console.print("\n[green]✅ All tests completed successfully![/green]")
        console.print("\n[cyan]Primary usage as a module:[/cyan]")
        console.print("from create_report_card_data import create_report_card_data")
        console.print("from retrieve_virus_data import retrieve_all_virus_data")
        console.print("")
        console.print("raw_df = retrieve_all_virus_data()")
        console.print("report_df = create_report_card_data(raw_df)")
        console.print("")
        console.print("[cyan]For fast testing:[/cyan]")
        console.print("test_df = create_report_card_data(raw_df, test_filter=True)  # Columbia, MO June 2024 only")
        
        return 0
        
    except Exception as e:
        console.print(f"[red]❌ Error: {e}[/red]")
        return 1


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)